@@ -33,6 +33,7 @@
 #include "tile.h"
 #include "tilelayer.h"
 #include "tileset.h"
+#include "objectgroup.h"
 
 #include <QtMath>
 
@@ -86,8 +87,9 @@ QRect IsometricRenderer::boundingRect(const QRect &rect) const
 QRectF IsometricRenderer::boundingRect(const MapObject *object) const
 {
     if (object->shape() == MapObject::Text) {
-        const QPointF topLeft = pixelToScreenCoords(object->position());
-        return QRectF(topLeft, object->size());
+        QRectF bounds { pixelToScreenCoords(object->position()), object->size() };
+        bounds.translate(-alignmentOffset(bounds, object->alignment(map())));
+        return bounds;
     } else if (object->shape() == MapObject::Point) {
         const qreal extraSpace = qMax(objectLineWidth() / 2, qreal(1));
         return shape(object).boundingRect()
@@ -96,25 +98,24 @@ QRectF IsometricRenderer::boundingRect(const MapObject *object) const
                                       extraSpace,
                                       extraSpace);
     } else if (!object->cell().isEmpty()) {
-        const QSizeF objectSize { object->size() };
-
-        QSizeF scale { 1.0, 1.0 };
-        QPoint tileOffset;
+        QRectF bounds { pixelToScreenCoords(object->position()), object->size() };
+        bounds.translate(-alignmentOffset(bounds, object->alignment(map())));
 
         if (const Tile *tile = object->cell().tile()) {
-            QSize imgSize = tile->size();
-            if (!imgSize.isNull()) {
-                scale = QSizeF(objectSize.width() / imgSize.width(),
-                               objectSize.height() / imgSize.height());
+            QPointF tileOffset = tile->offset();
+            const QSize tileSize = tile->size();
+            if (!tileSize.isNull()) {
+                const QSizeF scale {
+                    bounds.width() / tileSize.width(),
+                    bounds.height() / tileSize.height()
+                };
+                tileOffset.rx() *= scale.width();
+                tileOffset.ry() *= scale.height();
             }
-            tileOffset = tile->offset();
+            bounds.translate(tileOffset);
         }
 
-        const QPointF bottomCenter = pixelToScreenCoords(object->position());
-        return QRectF(bottomCenter.x() + (tileOffset.x() * scale.width()) - objectSize.width() / 2,
-                      bottomCenter.y() + (tileOffset.y() * scale.height()) - objectSize.height(),
-                      objectSize.width(),
-                      objectSize.height()).adjusted(-1, -1, 1, 1);
+        return bounds.adjusted(-1, -1, 1, 1);
     } else if (!object->polygon().isEmpty()) {
         qreal extraSpace = qMax(objectLineWidth(), qreal(1));
 
@@ -131,7 +132,10 @@ QRectF IsometricRenderer::boundingRect(const MapObject *object) const
     } else {
         // Take the bounding rect of the projected object, and then add a few
         // pixels on all sides to correct for the line width.
-        const QRectF base = pixelRectToScreenPolygon(object->bounds()).boundingRect();
+        QRectF bounds = object->bounds();
+        bounds.translate(-alignmentOffset(bounds, object->alignment(map())));
+
+        const QRectF base = pixelRectToScreenPolygon(bounds).boundingRect();
         const qreal extraSpace = qMax(objectLineWidth() / 2, qreal(1));
 
         return base.adjusted(-extraSpace,
@@ -146,12 +150,18 @@ QPainterPath IsometricRenderer::shape(const MapObject *object) const
 
     switch (object->shape()) {
     case MapObject::Ellipse: {
-        path.addEllipse(object->bounds());
+        QRectF bounds = object->bounds();
+        bounds.translate(-alignmentOffset(bounds, object->alignment(map())));
+
+        path.addEllipse(bounds);
         path = transform().map(path);
         break;
     }
     case MapObject::Rectangle: {
-        QPolygonF polygon = pixelRectToScreenPolygon(object->bounds());
+        QRectF bounds = object->bounds();
+        bounds.translate(-alignmentOffset(bounds, object->alignment(map())));
+
+        QPolygonF polygon = pixelRectToScreenPolygon(bounds);
         polygon.append(polygon.first());
         path.addPolygon(polygon);
         break;
@@ -185,9 +195,13 @@ QPainterPath IsometricRenderer::interactionShape(const MapObject *object) const
     } else {
         switch (object->shape()) {
         case MapObject::Rectangle:
-        case MapObject::Ellipse:
-            path.addPolygon(pixelRectToScreenPolygon(object->bounds()));
+        case MapObject::Ellipse: {
+            QRectF bounds = object->bounds();
+            bounds.translate(-alignmentOffset(bounds, object->alignment(map())));
+
+            path.addPolygon(pixelRectToScreenPolygon(bounds));
             break;
+        }
         case MapObject::Polygon:
         case MapObject::Text:
             path = shape(object);
@@ -251,6 +265,17 @@ void IsometricRenderer::drawGrid(QPainter *painter, const QRectF &rect,
 void IsometricRenderer::drawTileLayer(QPainter *painter,
                                       const TileLayer *layer,
                                       const QRectF &exposed) const
+{
+    CellRenderer renderer(painter, this, layer->effectiveTintColor());
+    auto tileRenderFunction = [&renderer](const Cell &cell, const QPointF &pos, const QSizeF &size) {
+        renderer.render(cell, pos, size, CellRenderer::BottomLeft);
+    };
+    drawTileLayer(layer, tileRenderFunction, exposed);
+}
+
+void IsometricRenderer::drawTileLayer(const TileLayer *layer,
+                                      const RenderTileCallback &renderTile,
+                                      const QRectF &exposed) const
 {
     const int tileWidth = map()->tileWidth();
     const int tileHeight = map()->tileHeight();
@@ -304,8 +329,6 @@ void IsometricRenderer::drawTileLayer(QPainter *painter,
     // Determine whether the current row is shifted half a tile to the right
     bool shifted = inUpperHalf ^ inLeftHalf;
 
-    CellRenderer renderer(painter, this);
-
     for (int y = startPos.y() * 2; y - tileHeight * 2 < rect.bottom() * 2;
          y += tileHeight)
     {
@@ -314,10 +337,9 @@ void IsometricRenderer::drawTileLayer(QPainter *painter,
         for (int x = startPos.x(); x < rect.right(); x += tileWidth) {
             const Cell &cell = layer->cellAt(columnItr);
             if (!cell.isEmpty()) {
-                Tile *tile = cell.tile();
-                QSize size = (tile && !tile->image().isNull()) ? tile->size() : map()->tileSize();
-                renderer.render(cell, QPointF(x, (qreal)y / 2), size,
-                                CellRenderer::BottomLeft);
+                const Tile *tile = cell.tile();
+                const QSize size = (tile && !tile->image().isNull()) ? tile->size() : map()->tileSize();
+                renderTile(cell, QPointF(x, (qreal)y / 2), size);
             }
 
             // Advance to the next column
@@ -343,8 +365,8 @@ void IsometricRenderer::drawTileSelection(QPainter *painter,
                                           const QColor &color,
                                           const QRectF &exposed) const
 {
-    painter->setBrush(color);
-    painter->setPen(Qt::NoPen);
+    QPainterPath path;
+
 #if QT_VERSION < 0x050800
     const auto rects = region.rects();
     for (const QRect &r : rects) {
@@ -353,8 +375,19 @@ void IsometricRenderer::drawTileSelection(QPainter *painter,
 #endif
         QPolygonF polygon = tileRectToScreenPolygon(r);
         if (QRectF(polygon.boundingRect()).intersects(exposed))
-            painter->drawConvexPolygon(polygon);
+            path.addPolygon(polygon);
     }
+
+    QColor penColor(color);
+    penColor.setAlpha(255);
+
+    QPen pen(penColor);
+    pen.setCosmetic(true);
+
+    painter->setPen(pen);
+    painter->setBrush(color);
+    painter->setRenderHint(QPainter::Antialiasing, true);
+    painter->drawPath(path.simplified());
 }
 
 void IsometricRenderer::drawMapObject(QPainter *painter,
@@ -369,49 +402,46 @@ void IsometricRenderer::drawMapObject(QPainter *painter,
     const Cell &cell = object->cell();
 
     if (!cell.isEmpty()) {
-        const QSizeF size = object->size();
-        const QPointF pos = pixelToScreenCoords(object->position());
+        QRectF bounds = { pixelToScreenCoords(object->position()), object->size() };
+        bounds.translate(-alignmentOffset(bounds, object->alignment(map())));
 
-        CellRenderer(painter, this).render(cell, pos, size,
-                                           CellRenderer::BottomCenter);
+        CellRenderer(painter, this, object->objectGroup()->effectiveTintColor())
+                .render(cell, bounds.topLeft(), bounds.size());
 
         if (testFlag(ShowTileObjectOutlines)) {
-            QPointF tileOffset;
-            QPointF scale(1.0, 1.0);
-
-            if (const Tile *tile = cell.tile()) {
-                tileOffset = tile->offset();
-
-                const QPixmap &image = tile->image();
-                const QSizeF imageSize = image.size();
-
-                if (!imageSize.isEmpty()) {
-                    scale = QPointF(size.width() / imageSize.width(),
-                                    size.height() / imageSize.height());
+            if (const Tile *tile = object->cell().tile()) {
+                QPointF tileOffset = tile->offset();
+                const QSize tileSize = tile->size();
+                if (!tileSize.isNull()) {
+                    const QSizeF scale {
+                        bounds.width() / tileSize.width(),
+                        bounds.height() / tileSize.height()
+                    };
+                    tileOffset.rx() *= scale.width();
+                    tileOffset.ry() *= scale.height();
                 }
+                bounds.translate(tileOffset);
             }
 
-            QRectF rect(QPointF(pos.x() - size.width() / 2 + tileOffset.x() * scale.x(),
-                                pos.y() - size.height() + tileOffset.y() * scale.y()),
-                        size);
-
             pen.setStyle(Qt::SolidLine);
             painter->setRenderHint(QPainter::Antialiasing, false);
             painter->setBrush(Qt::NoBrush);
             painter->setPen(pen);
-            painter->drawRect(rect);
+            painter->drawRect(bounds);
             pen.setStyle(Qt::DotLine);
             pen.setColor(color);
             painter->setPen(pen);
-            painter->drawRect(rect);
+            painter->drawRect(bounds);
         }
     } else if (object->shape() == MapObject::Text) {
-        const QPointF pos = pixelToScreenCoords(object->position());
+        QRectF bounds = { pixelToScreenCoords(object->position()), object->size() };
+        bounds.translate(-alignmentOffset(bounds, object->alignment(map())));
+
         const auto& textData = object->textData();
 
         painter->setFont(textData.font);
         painter->setPen(textData.color);
-        painter->drawText(QRectF(pos, object->size()),
+        painter->drawText(bounds,
                           textData.text,
                           textData.textOption());
     } else {
@@ -433,11 +463,14 @@ void IsometricRenderer::drawMapObject(QPainter *painter,
         painter->setPen(pen);
         painter->setRenderHint(QPainter::Antialiasing);
 
+        QRectF bounds = object->bounds();
+        bounds.translate(-alignmentOffset(bounds, object->alignment(map())));
+
         // TODO: Do something sensible to make null-sized objects usable
 
         switch (object->shape()) {
         case MapObject::Ellipse: {
-            const QPolygonF rect = pixelRectToScreenPolygon(object->bounds());
+            const QPolygonF rect = pixelRectToScreenPolygon(bounds);
             const QPainterPath ellipse = shape(object);
 
             painter->drawPath(ellipse);
@@ -457,7 +490,7 @@ void IsometricRenderer::drawMapObject(QPainter *painter,
             drawPointObject(painter, color);
             break;
         case MapObject::Rectangle: {
-            QPolygonF polygon = pixelRectToScreenPolygon(object->bounds());
+            QPolygonF polygon = pixelRectToScreenPolygon(bounds);
             painter->drawPolygon(polygon);
 
             painter->setPen(colorPen);
