@@ -24,26 +24,27 @@
 #include "documentmanager.h"
 #include "imagecolorpickerwidget.h"
 #include "mapdocument.h"
-#include "preferences.h"
+#include "session.h"
 #include "utils.h"
 
 #include <QFileDialog>
 #include <QFileInfo>
 #include <QImage>
 #include <QMessageBox>
-#include <QSettings>
 #include <QCheckBox>
 
-static const char * const TYPE_KEY = "Tileset/Type";
-static const char * const EMBED_KEY = "Tileset/EmbedInMap";
-static const char * const COLOR_ENABLED_KEY = "Tileset/UseTransparentColor";
-static const char * const COLOR_KEY = "Tileset/TransparentColor";
-static const char * const TILE_SIZE_KEY = "Tileset/TileSize";
-static const char * const SPACING_KEY = "Tileset/Spacing";
-static const char * const MARGIN_KEY = "Tileset/Margin";
-
 using namespace Tiled;
 
+namespace session {
+static SessionOption<int> tilesetType { "tileset.type" };
+static SessionOption<bool> embedInMap { "tileset.embedInMap" };
+static SessionOption<bool> useTransparentColor { "tileset.useTransparentColor" };
+static SessionOption<QColor> transparentColor { "tileset.transparentColor", Qt::magenta };
+static SessionOption<QSize> tileSize { "tileset.tileSize", QSize(32, 32) };
+static SessionOption<int> tilesetSpacing { "tileset.spacing" };
+static SessionOption<int> tilesetMargin { "tileset.margin" };
+} // namespace session
+
 enum TilesetType {
     TilesetImage,
     ImageCollection
@@ -70,28 +71,16 @@ NewTilesetDialog::NewTilesetDialog(QWidget *parent) :
     setWindowFlags(windowFlags() & ~Qt::WindowContextHelpButtonHint);
 #endif
 
-    // Restore previously used settings
-    QSettings *s = Preferences::instance()->settings();
+    const QSize tileSize = session::tileSize;
 
-    int tilesetType = s->value(QLatin1String(TYPE_KEY)).toInt();
-    bool embedded = s->value(QLatin1String(EMBED_KEY)).toBool();
-    bool colorEnabled = s->value(QLatin1String(COLOR_ENABLED_KEY)).toBool();
-    QString colorName = s->value(QLatin1String(COLOR_KEY)).toString();
-    QColor color = QColor::isValidColor(colorName) ? QColor(colorName) : Qt::magenta;
-    QSize tileSize = s->value(QLatin1String(TILE_SIZE_KEY)).toSize();
-    int spacing = s->value(QLatin1String(SPACING_KEY)).toInt();
-    int margin = s->value(QLatin1String(MARGIN_KEY)).toInt();
-
-    mUi->tilesetType->setCurrentIndex(tilesetType);
-    mUi->embedded->setChecked(embedded);
-    mUi->useTransparentColor->setChecked(colorEnabled);
-    mUi->colorButton->setColor(color);
-    if (tileSize.isValid()) {
-        mUi->tileWidth->setValue(tileSize.width());
-        mUi->tileHeight->setValue(tileSize.height());
-    }
-    mUi->spacing->setValue(spacing);
-    mUi->margin->setValue(margin);
+    mUi->tilesetType->setCurrentIndex(session::tilesetType);
+    mUi->embedded->setChecked(session::embedInMap);
+    mUi->useTransparentColor->setChecked(session::useTransparentColor);
+    mUi->colorButton->setColor(session::transparentColor);
+    mUi->tileWidth->setValue(tileSize.width());
+    mUi->tileHeight->setValue(tileSize.height());
+    mUi->spacing->setValue(session::tilesetSpacing);
+    mUi->margin->setValue(session::tilesetMargin);
 
     connect(mUi->browseButton, &QAbstractButton::clicked, this, &NewTilesetDialog::browse);
     connect(mUi->name, &QLineEdit::textEdited, this, &NewTilesetDialog::nameEdited);
@@ -107,7 +96,7 @@ NewTilesetDialog::NewTilesetDialog(QWidget *parent) :
     connect(mUi->buttonBox, &QDialogButtonBox::accepted, this, &NewTilesetDialog::tryAccept);
     connect(mUi->buttonBox, &QDialogButtonBox::rejected, this, &NewTilesetDialog::reject);
 
-    mUi->imageGroupBox->setVisible(tilesetType == 0);
+    mUi->imageGroupBox->setVisible(session::tilesetType == 0);
     updateOkButton();
 }
 
@@ -199,9 +188,6 @@ bool NewTilesetDialog::editTilesetParameters(TilesetParameters &parameters)
 
 void NewTilesetDialog::tryAccept()
 {
-    // Used for storing the settings for next time
-    QSettings *s = Preferences::instance()->settings();
-
     const QString name = mUi->name->text();
 
     SharedTileset tileset;
@@ -242,19 +228,19 @@ void NewTilesetDialog::tryAccept()
         }
 
         if (mMode == CreateTileset) {
-            s->setValue(QLatin1String(COLOR_ENABLED_KEY), useTransparentColor);
-            s->setValue(QLatin1String(COLOR_KEY), transparentColor.name());
-            s->setValue(QLatin1String(TILE_SIZE_KEY), QSize(tileWidth, tileHeight));
-            s->setValue(QLatin1String(SPACING_KEY), spacing);
-            s->setValue(QLatin1String(MARGIN_KEY), margin);
+            session::useTransparentColor = useTransparentColor;
+            session::transparentColor = transparentColor;
+            session::tileSize = QSize(tileWidth, tileHeight);
+            session::tilesetSpacing = spacing;
+            session::tilesetMargin = margin;
         }
     } else {
         tileset = Tileset::create(name, 1, 1);
     }
 
     if (mMode == CreateTileset) {
-        s->setValue(QLatin1String(TYPE_KEY), mUi->tilesetType->currentIndex());
-        s->setValue(QLatin1String(EMBED_KEY), mUi->embedded->isChecked());
+        session::tilesetType = mUi->tilesetType->currentIndex();
+        session::embedInMap = mUi->embedded->isChecked();
     }
 
     mNewTileset = tileset;
